# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Use ply library to translate between Portables and a text representation.

Part of the upgrade to allow passing of any type, e.g. including compositions of
list, dict etc, rather than exclusively registered messages.

A portable type is presented to portable_to_signature, e.g. MapOf(Unicode(),Person)
which generates a string like "map<unicode,module.Person>". Presenting that same
text to signature_to_portable will recover the semantic equal of the original, i.e.
a MapOf(etc...).
"""

import sys
from .virtual_memory import *
from enum import Enum

__all__ = [
	'signature_to_portable',
	'portable_to_signature',
	'portable_to_tag',
]

NAME_CLASS = {
	'boolean': Boolean,
	'integer2': Integer2,
	'integer4': Integer4,
	'integer8': Integer8,
	'unsigned2': Unsigned2,
	'unsigned4': Unsigned4,
	'unsigned8': Unsigned8,
	'float4': Float4,
	'float8': Float8,
	'byte': Byte,
	'block': Block,
	'character': Character,
	'string': String,
	'rune': Rune,
	'unicode': Unicode,
	'clock': ClockTime,
	'span': TimeSpan,
	'world': WorldTime,
	'delta': TimeDelta,
	'uuid': UUID,
	'any': Any,
	'target': TargetAddress,
	'address': Address,
	'type': Type,
	'word': Word,
	'array': ArrayOf,
	'vector': VectorOf,
	'deque': DequeOf,
	'set': SetOf,
	'map': MapOf,
	'pointer': PointerTo,
}

tokens = (
	'NUMBER',
	'NAME',
	'LEFT', 'SEPARATOR', 'RIGHT',
)

# Tokens
t_LEFT = r'<'
t_SEPARATOR = r','
t_RIGHT = r'>'

def t_NUMBER(t):
	r'\d+'
	try:
		value = int(t.value)
	except ValueError:
		value = 0
	t.value = value
	return t

class NotFound(object): pass

def sys_module_class(s):
	"""Convert the dotted string *s* to a class, or None."""
	try:
		i = s.rindex('.')
	except ValueError:
		return None
	module = s[:i]
	name = s[i + 1:]
	try:
		m = sys.modules[module]
	except KeyError:
		return None

	try:
		c = m.__dict__[name]
	except KeyError:
		return None
	return c

def t_NAME(t):
	r'[_a-zA-Z][_a-zA-Z0-9]*(\.[_a-zA-Z0-9]+)*'
	#
	c = NAME_CLASS.get(t.value, NotFound)
	if c is NotFound:
		c = sys_module_class(t.value)
		if c is None:
			raise ValueError(f'unknown class {t.value}')
	t.value = c
	return t

# Ignored characters
t_ignore = " \t"

def t_newline(t):
	r'\n+'
	t.lexer.lineno += t.value.count("\n")

def t_error(t):
	token = t.value[0]
	e = f'Illegal character "{token}"'
	t.lexer.skip(1)

# Build the lexer
import ply.lex as lex
lexer = lex.lex()

# Parsing rules

precedence = ()

def p_statement_type(t):
	'''statement : type'''
	t[0] = t[1]

def p_type_name(t):
	'''type : NAME'''
	c = t[1]
	if is_portable_class(c):
		if is_container_class(c):
			raise ValueError('container class and no parameters')
		t[0] = c()
	elif issubclass(c, Enum):
		t[0] = Enumeration(c)
	else:
		t[0] = UserDefined(c)

def p_type_name_list(t):
	'''type : NAME LEFT list RIGHT'''
	c = t[1]
	a = t[3]
	if not is_container_class(c):
		raise ValueError('non-container class with parameters')
	if c in (MapOf, ArrayOf):
		t[0] = c(a[0], a[1])
	else:
		t[0] = c(a[0])

def p_list_arg(t):
	'list : arg'
	t[0] = [t[1]]

def p_list_list_arg(t):
	'list : list SEPARATOR arg'
	tl = t[1]
	ta = t[3]
	tl.append(ta)
	t[0] = tl

def p_arg_NUMBER(t):
	'arg : NUMBER'
	number = int(t[1])
	t[0] = number

def p_arg_type(t):
	'arg : type'
	t[0] = t[1]

def p_error(t):
	# self_service.fault('syntax error at "{error}"'.format(error=t.value))
	pass

import ply.yacc as yacc
parser = yacc.yacc(debug=False)

#
#
def signature_to_portable(text):
	'''Given a unique string representation, generate the associated type. Return a Portable.'''
	t = parser.parse(text)
	return t

#
#
CLASS_NAME = {
	Boolean: 'boolean',
	Integer2: 'integer2',
	Integer4: 'integer4',
	Integer8: 'integer8',
	Unsigned2: 'unsigned2',
	Unsigned4: 'unsigned4',
	Unsigned8: 'unsigned8',
	Float4: 'float4',
	Float8: 'float8',
	Byte: 'byte',
	Block: 'block',
	Character: 'character',
	String: 'string',
	Rune: 'rune',
	Unicode: 'unicode',
	ClockTime: 'clock',
	TimeSpan: 'span',
	WorldTime: 'world',
	TimeDelta: 'delta',
	UUID: 'uuid',
	Enumeration: 'not used',
	Any: 'any',
	TargetAddress: 'target',
	Address: 'address',
	Type: 'type',
	Word: 'word',
	ArrayOf: 'array',
	VectorOf: 'vector',
	DequeOf: 'deque',
	SetOf: 'set',
	MapOf: 'map',
	UserDefined: 'not used',
	PointerTo: 'pointer',
}

#
#
def portable_to_signature(a):
	'''Given a Portable, generate a unique signature. Return a string.'''
	c = a.__class__
	t = CLASS_NAME.get(c, NotFound)
	if t is NotFound:
		return ''

	if c in (UserDefined, Enumeration):
		m = a.element.__module__
		s = a.element.__name__
		return f'{m}.{s}'
	if c in (VectorOf, DequeOf, SetOf, PointerTo):
		e = portable_to_signature(a.element)
		return f'{t}<{e}>'
	if c == MapOf:
		k = portable_to_signature(a.key)
		v = portable_to_signature(a.value)
		return f'{t}<{k},{v}>'
	if c == ArrayOf:
		e = portable_to_signature(a.element)
		return f'{t}<{e},{a.size}>'

	return t

TAG_NAME = {
	Boolean: 'bool',
	Integer2: 'integer2',
	Integer4: 'integer4',
	Integer8: 'int',
	Unsigned2: 'unsigned2',
	Unsigned4: 'unsigned4',
	Unsigned8: 'unsigned8',
	Float4: 'float4',
	Float8: 'float',
	Byte: 'byte',
	Block: 'bytearray',
	Character: 'character',
	String: 'bytes',
	Rune: 'rune',
	Unicode: 'str',
	ClockTime: 'clock',
	TimeSpan: 'span',
	WorldTime: 'datetime',
	TimeDelta: 'timedelta',
	UUID: 'UUID',
	Enumeration: 'not used',
	Any: 'any',
	TargetAddress: 'target',
	Address: 'address',
	Type: 'type',
	Word: 'word',
	ArrayOf: 'array',
	VectorOf: 'list',
	DequeOf: 'deque',
	SetOf: 'set',
	MapOf: 'dict',
	UserDefined: 'not used',
	PointerTo: 'pointer',
}

def portable_to_tag(a):
	'''Given a Portable, generate a unique signature. Return a string.'''
	c = a.__class__
	t = TAG_NAME.get(c, NotFound)
	if t is NotFound:
		return ''

	if c == UserDefined:
		return a.element.__art__.name
	if c == Enumeration:
		m = a.element.__module__
		s = a.element.__name__
		return f'{m}_{s}'
	if c in (VectorOf, DequeOf, SetOf, PointerTo):
		e = portable_to_tag(a.element)
		return f'{t}_{e}'
	if c == MapOf:
		k = portable_to_tag(a.key)
		v = portable_to_tag(a.value)
		return f'dict_{k}_{v}'
	if c == ArrayOf:
		e = portable_to_signature(a.element)
		return f'array_{e}_{a.size}'

	return t



# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2017-2023 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Movement of files to home processes and back.

"""
__docformat__ = 'restructuredtext'

import os
import uuid
import shutil
from datetime import datetime

from .bind_type import *
from .virtual_memory import *
from .message_memory import *
from .convert_memory import *
from .virtual_codec import *
from .json_codec import *
from .file_object import *
from .folder_object import *
from .point_runtime import *
from .virtual_point import *
from .point_machine import *

__all__ = [
	'DELTA_FILE_ADD', 'DELTA_FILE_UPDATE', 'DELTA_FILE_UGM', 'DELTA_FILE_REMOVE',
	'DELTA_FOLDER_ADD', 'DELTA_FOLDER_UPDATE', 'DELTA_FOLDER_UGM', 'DELTA_FOLDER_REMOVE',
	'DELTA_FILE_CRUD', 'DELTA_FOLDER_CRUD',
	'DELTA_CRUD',
	'StorageTables',
	'StorageAttributes',
	'StorageManifest',
	'StorageListing',
	'StorageTables',
	'storage_manifest',
	'storage_selection',
	'storage_walk',
	'show_listings',
	'storage_delta',
	'DeltaMachine',
	'RemoveFolder',
	'RemoveFile',
	'UpdateFile',
	'UpdateUser',
	'UpdateGroup',
	'UpdateMode',
	'AddFile',
	'AddFolder',
	'ReplaceWithFile',
	'ReplaceWithFolder',
	'FolderTransfer',
]

#
#
class INITIAL: pass
class RUNNING: pass
class HALTED: pass

# User, group ids and mode (permissions - rwxrwxrwx).
#
class StorageAttributes(object):
	"""Attributes of a folder/file in a storage hierarchy."""
	def __init__(self, user: int=None, group: int=None, mode: int=None):
		self.user = user
		self.group = group
		self.mode = mode

	def __str__(self):
	# user = pwd.getpwuid(self.user)
	# group = grp.getgrgid(self.group)
	# rwx = stat.filemode(self.mode)
	# return f'StorageAttributes(owner={user.pw_name}, group={group.gr_name}, permissions={rwx})'
		return f'StorageAttributes(owner={self.user}, group={self.group}, permissions={self.mode})'

bind(StorageAttributes, copy_before_sending=False)

#
#
class StorageManifest(object):
	"""Folder in a folder/file storage hierarchy."""
	def __init__(self, path: str=None,
			modified: datetime=None, attributes: StorageAttributes=None,
			content: dict[str, Any]=None, parent=None,
			manifests: int=0, listings: int=0, bytes: int=0):
		self.path = path
		self.modified = modified
		self.attributes = attributes or StorageAttributes()
		self.content = content or {}
		self.parent = parent
		self.manifests = manifests
		self.listings = listings
		self.bytes = bytes

	def full_path(self):
		return self.path

	def origin(self):
		if self.parent is None:
			return self.path
		return self.parent.origin()

	def __str__(self):
		return f'StorageManifest(path={self.path}, modified={self.modified}, attributes={self.attributes}, contents={len(self.content)})'

bind(StorageManifest, copy_before_sending=False, parent=PointerTo(StorageManifest))

#
#
class StorageListing(object):
	"""File in a folder/file storage hierarchy."""
	def __init__(self, name: str=None,
			modified: datetime=None, attributes: StorageAttributes=None,
			size: int=0, parent=None):
		self.name = name
		self.modified = modified
		self.attributes = attributes or StorageAttributes()
		self.size = size
		self.parent = parent

	def full_path(self):
		p = os.path.join(self.parent.path, self.name)
		return p

	def origin(self):
		return self.parent.origin()

	def __str__(self):
		return f'StorageListing(name={self.name}, modified={self.modified}, attributes={self.attributes})'

bind(StorageListing, copy_before_sending=False, parent=PointerTo(StorageManifest))

#
#
class StorageTables(object):
	"""Names for ids appearing in a folder/file hierarchy."""
	def __init__(self, user_name: dict[int, str]=None, group_name: dict[int, str]=None):
		self.user_name = user_name or {}
		self.group_name = group_name or {}

	def __str__(self):
		return f'StorageTables(user_name={len(self.user_name)}, group_name={len(self.group_name)})'

bind(StorageTables, copy_before_sending=False)

#
#
def storage_manifest(path, parent=None, table=None):
	"""Scan the given folder, recursively. Return a manifest of contents."""
	table = table or StorageTables()
	user_name = table.user_name
	group_name = table.group_name

	# Small optimization to reduce hits on the
	# platform dbs.
	def lookup(s):
		# User ids <---> names
		try:
			t = user_name[s.st_uid]	 # Already seen?
		except KeyError:
			t = None				# Not defined.
			#try:
			#	t = pwd.getpwuid(s.st_uid).pw_name	# Nope - lookup.
			#except KeyError:
			#	t = None				# Not defined.
			user_name[s.st_uid] = t	 # Remember results.
		# Group ids <---> names
		try:
			t = group_name[s.st_gid]
		except KeyError:
			t = None
			#try:
			#	t = grp.getgrgid(s.st_gid).gr_name
			#except KeyError:
			#	t = None
			group_name[s.st_gid] = t
		a = StorageAttributes(user=s.st_uid, group=s.st_gid, mode=s.st_mode)
		return a

	s = os.stat(path)
	d = datetime.fromtimestamp(s.st_mtime, tz=UTC)
	a = lookup(s)
	m = StorageManifest(path=path, modified=d, attributes=a, parent=parent)

	content = m.content
	for k in os.listdir(path):
		p = os.path.join(path, k)
		if os.path.isdir(p):
			m.manifests += 1
			v, _ = storage_manifest(p, parent=m, table=table)
			m.manifests += v.manifests
			m.listings += v.listings
			m.bytes += v.bytes
		elif os.path.isfile(p):
			m.listings += 1
			s = os.stat(p)
			d = datetime.fromtimestamp(s.st_mtime, tz=UTC)
			a = lookup(s)
			v = StorageListing(name=k, modified=d, attributes=a, size=s.st_size, parent=m)
			m.bytes += v.size
		else:
			continue
		content[k] = v

	return m, table

#
def storage_selection(selection, path=None, table=None):
	"""Gather arbitrary files and folders into a single logical manifest. Return a manifest of contents."""
	table = table or StorageTables()
	user_name = table.user_name
	group_name = table.group_name

	# Small optimization to reduce hits on the
	# platform dbs.
	def lookup(s):
		# User ids <---> names
		try:
			t = user_name[s.st_uid]	 # Already seen?
		except KeyError:
			t = None				# Not defined.
			#try:
			#	t = pwd.getpwuid(s.st_uid).pw_name	# Nope - lookup.
			#except KeyError:
			#	t = None				# Not defined.
			user_name[s.st_uid] = t	 # Remember results.
		# Group ids <---> names
		try:
			t = group_name[s.st_gid]
		except KeyError:
			t = None
			#try:
			#	t = grp.getgrgid(s.st_gid).gr_name
			#except KeyError:
			#	t = None
			group_name[s.st_gid] = t
		a = StorageAttributes(user=s.st_uid, group=s.st_gid, mode=s.st_mode)
		return a

	selected = StorageManifest(path=path)	# This is a purely logical collection, i.e. path is optional.
	parenting = {}							# May include materials from multiple sources.

	def find(parent):
		p = parenting.get(parent, None)
		if p is None:
			s = os.stat(parent)
			d = datetime.fromtimestamp(s.st_mtime, tz=UTC)
			a = lookup(s)
			p = StorageManifest(path=parent, modified=d, attributes=a)
			parenting[parent] = p
		return p

	content = selected.content
	for s in selection:
		absolute = os.path.abspath(s)
		if not os.path.exists(absolute):
			raise ValueError(f'"{absolute}" does not exist')
		s0, s1 = os.path.split(absolute)
		if s1 in content:
			raise ValueError(f'multiple selections of "{s1}"')
		parent = find(s0)

		if os.path.isdir(absolute):
			selected.manifests += 1
			v, _ = storage_manifest(path=absolute, parent=parent, table=table)
			selected.manifests += v.manifests
			selected.listings += v.listings
			selected.bytes += v.bytes
		elif os.path.isfile(absolute):
			selected.listings += 1
			s = os.stat(absolute)
			d = datetime.fromtimestamp(s.st_mtime, tz=UTC)
			a = lookup(s)
			v = StorageListing(name=s1, modified=d, attributes=a, size=s.st_size, parent=parent)
			selected.bytes += v.size
		content[s1] = v

	return selected, table

# Flatten the hierarchy into a list
# of full name and manifest/listing.
def storage_walk(man):
	for k, v in man.content.items():
		if isinstance(v, StorageManifest):
			yield from storage_walk(v)
		elif isinstance(v, StorageListing):
			yield v.full_path(), v
	yield man.path, man

# Printing of a storage area.
#
def show_file(header, listing):
	print('%s- %s (%d)' % (header, listing.name, listing.size))

def show_folder(header, manifest):
	p, f = os.path.split(manifest.path)
	print('%s+ %s (%d, %d, %d)' % (header, f, manifest.manifests, manifest.listings, manifest.bytes))

def show_listings(man):
	path, name = os.path.split(man.path)
	c = len(path) + 1

	d = {k[c:]: v for k, v in storage_walk(man)}
	s = sorted(d)
	for p in s:
		v = d[p]
		t = p.split(os.path.sep)
		n = len(t)
		if n == 1:
			print('%s (%d, %d, %d)' % (man.path, man.manifests, man.listings, man.bytes))
		elif isinstance(v, StorageManifest):
			h = '  ' * (n - 2)
			show_folder(h, v)
		else:
			h = '  ' * (n - 2)
			show_file(h, v)

class TransferHalted(Exception):
	pass

DELTA_FILE_ADD = 0x01
DELTA_FILE_UPDATE = 0x02
DELTA_FILE_UGM = 0x04
DELTA_FILE_REMOVE = 0x08
DELTA_FILE_CRUD = DELTA_FILE_ADD | DELTA_FILE_UPDATE | DELTA_FILE_UGM | DELTA_FILE_REMOVE

DELTA_FOLDER_ADD = 0x10
DELTA_FOLDER_UPDATE = 0x20		# Walk down into sub-folders.
DELTA_FOLDER_UGM = 0x40
DELTA_FOLDER_REMOVE = 0x80
DELTA_FOLDER_CRUD = DELTA_FOLDER_ADD | DELTA_FOLDER_UPDATE | DELTA_FOLDER_UGM | DELTA_FOLDER_REMOVE

DELTA_CRUD = DELTA_FILE_CRUD | DELTA_FOLDER_CRUD

class DeltaMachine(object):
	def __init__(self, context, target):
		self.context = context
		self.target = target
		self.alias_file = {}
		self.alias_folder = {}

	# Map a source to its final target.
	def resting(self, ml):
		origin = ml.origin()	  # Location of source.
		path = ml.full_path()		# Full path and name.
		len_origin = len(origin)

		source, name = os.path.split(path)  # Clip name off.
		if len_origin == len(source):
			t = os.path.join(self.target, name)	# No relative component.
		else:
			t = os.path.join(self.target, source[len_origin + 1:], name)
		return t

	def alias(self, d, m, t=None):
		s = str(uuid.uuid4())		   # Actual alias.
		origin = d.source.origin()	  # Location of source.
		len_origin = len(origin)

		path = d.source.full_path()
		source, _ = os.path.split(path)  # Clip name off.
		# Want to replace the source location with the target
		# location and the final name with the alias name.
		# Extra detail is including the relative path, if any.
		if t is not None:
			t = os.path.join(t, s)	# No relative component.
		elif len_origin == len(source):
			t = os.path.join(self.target, s)	# No relative component.
		else:
			t = os.path.join(self.target, source[len_origin + 1:], s)
		m[path] = t
		return t

	def aliases(self):
		a = len(self.alias_file) + len(self.alias_folder)
		return a

	def rename(self):
		for k, a in self.alias_file.items():
			#if self.halted:
			#	return Aborted()
			p, _ = os.path.split(a)
			_, t = os.path.split(k)
			e = os.path.join(p, t)
			os.rename(a, e)
		self.alias_file = {}
		for k, a in self.alias_folder.items():
			#if self.halted:
			#	return Aborted()
			p, _ = os.path.split(a)
			_, t = os.path.split(k)
			e = os.path.join(p, t)
			os.rename(a, e)
		self.alias_folder = {}

	def clear(self):
		"""Clean up any transfer artefacts."""
		for _, v in self.alias_file.items():
			try:
				os.remove(v)
			except FileNotFoundError:
				pass

		for _, v in self.alias_folder.items():
			try:
				remove_folder(v)
			except FileNotFoundError:
				pass

# Actual file and folder copying fragements.
def transfer_file(self, source, target):
	"""Copy a single executable to an interim alias."""
	with open(source, 'rb') as r, open(target, 'wb') as w:
		while True:
			if self.halted:
				raise TransferHalted()
			t = r.read(4096)
			if t == b'':
				break
			w.write(t)
	shutil.copystat(source, target)

def add_folder(self, source, target):
	"""Add folder cos folders are never transferred, i.e. like a file update."""
	try:
		os.makedirs(target)
		shutil.copystat(source, target)
	except FileExistsError:		# Might be hiding a bug with this.
		pass
	# Loop through contents, as defined by the
	# source folder.
	for f in os.listdir(source):
		if self.halted:
			raise TransferHalted()
		s = os.path.join(source, f)
		t = os.path.join(target, f)
		if os.path.isdir(s):
			add_folder(self, s, t)
		elif os.path.isfile(s):
			transfer_file(self, s, t)

#
#
class RemoveFolder(object):
	def __init__(self, folder=None):
		self.folder = folder

	def __str__(self):
		return f'RemoveFolder(path={self.folder.path}, modified={self.folder.modified}, contents={len(self.folder.content)})'

	def __call__(self, dm):
		p = self.folder.path
		remove_folder(p)

bind(RemoveFolder, folder=PointerTo(StorageManifest))


class RemoveFile(object):
	def __init__(self, file=None):
		self.file = file

	def __str__(self):
		return f'RemoveFile(path={self.file.full_path()}, modified={self.file.modified})'

	def __call__(self, dm):
		f = self.file.full_path()
		os.remove(f)

bind(RemoveFile, file=PointerTo(StorageListing))


class UpdateFile(object):
	def __init__(self, source=None, target=None):
		self.source = source
		self.target = target

	def __str__(self):
		return f'UpdateFile(path={self.source.full_path()}, modified={self.source.modified}, target={self.target.parent.path})'

	def __call__(self, dm):
		a = dm.alias(self, dm.alias_file, t=self.target.parent.path)
		transfer_file(dm.context, self.source.full_path(), a)

bind(UpdateFile, source=PointerTo(StorageListing), target=PointerTo(StorageListing))


class UpdateUser(object):
	def __init__(self, target=None, user: int=None):
		self.target = target
		self.user = user

	def __str__(self):
		return f'UpdateUser(target={self.target.full_path()}, user={self.user})'

	def __call__(self, dm):
		pass

bind(UpdateUser, target=PointerTo(Any()))

class UpdateGroup(object):
	def __init__(self, target=None, group: int=None):
		self.target = target
		self.group = group

	def __str__(self):
		return f'UpdateGroup(target={self.target.full_path()}, group={self.group})'

	def __call__(self, dm):
		pass

bind(UpdateGroup, target=PointerTo(Any()))


class UpdateMode(object):
	def __init__(self, target=None, mode: int=None):
		self.target = target
		self.mode = mode

	def __str__(self):
		return f'UpdateMode(target={self.target.full_path()}, mode={self.mode})'

	def __call__(self, dm):
		f = self.target.full_path()
		os.chmod(f, self.mode)

bind(UpdateMode, target=PointerTo(Any()))


class AddFile(object):
	def __init__(self, source=None, target=None):
		self.source = source
		self.target = target

	def __str__(self):
		return f'AddFile(path={self.source.full_path()}, target={self.target.path})'

	def __call__(self, dm):
		a = dm.alias(self, dm.alias_file, t=self.target.path)
		transfer_file(dm.context, self.source.full_path(), a)

bind(AddFile, source=PointerTo(StorageListing), target=PointerTo(StorageManifest))

class AddFolder(object):
	def __init__(self, source=None, target=None):
		self.source = source
		self.target = target

	def __str__(self):
		return f'AddFolder(path={self.source.path}, target={self.target.path})'

	def __call__(self, dm):
		a = dm.alias(self, dm.alias_folder, t=self.target.path)
		add_folder(dm.context, self.source.path, a)

bind(AddFolder, source=PointerTo(StorageManifest), target=PointerTo(StorageManifest))

class ReplaceWithFile(object):
	def __init__(self, source=None, target=None):
		self.source = source
		self.target = target

	def __str__(self):
		return f'ReplaceWithFile(source={self.source.full_path()}, target={self.target.path})'

	def __call__(self, dm):
		p = self.target.path
		remove_folder(p)
		a = dm.alias(self, dm.alias_file)
		transfer_file(dm.context, self.source.full_path(), a)

bind(ReplaceWithFile, source=PointerTo(StorageListing), target=PointerTo(StorageManifest))

class ReplaceWithFolder(object):
	def __init__(self, source=None, target=None):
		self.source = source
		self.target = target

	def __str__(self):
		return f'ReplaceWithFolder(path={self.source.path}, target={self.target.full_path()})'

	def __call__(self, dm):
		f = self.target.full_path()
		os.remove(f)
		a = dm.alias(self, dm.alias_folder)
		add_folder(self, self.source.path, a)

bind(ReplaceWithFolder, source=PointerTo(StorageManifest), target=PointerTo(StorageListing))

#
#
def storage_delta(source, target, flags=DELTA_CRUD, guard_path=None):
	"""Compare the two trees. Yield a sequence of the changes needed."""
	def content_keys(m):
		if m is None:
			return []
		return m.content.keys()
	keys = content_keys(source) | content_keys(target)

	# Source and target guaranteed to exist and represent the
	# equivalent nodes (i.e. manifests) of their respective trees.
	# Iterate through this pair of nodes using the union of the two
	# sets of keys.

	for k in keys:
		s = None if source is None else source.content.get(k, None)
		t = None if target is None else target.content.get(k, None)
		# Possibilities are;
		# s, t ...... compare
		# s, None ... add
		# None, t ... remove
		# (Note: None, None never happens)
		if s:
			if t:
				# At equivalent sub-nodes in the tree.
				# Could be 2 folders or 2 files, or
				# mismatched folder and file.
				if type(s) == type(t):
					if isinstance(s, StorageManifest):
						if flags & DELTA_FOLDER_UPDATE: yield from storage_delta(s, t, flags=flags)			# Recurse.
						if s.attributes.user != t.attributes.user and flags & DELTA_FOLDER_UGM:
							yield UpdateUser(t, s.attributes.user)
						if s.attributes.group != t.attributes.group and flags & DELTA_FOLDER_UGM:
							yield UpdateGroup(t, s.attributes.group)
						if s.attributes.mode != t.attributes.mode and flags & DELTA_FOLDER_UGM:
							yield UpdateMode(t, s.attributes.mode)
					else:
						if s.modified > t.modified and flags & DELTA_FILE_UPDATE:
							yield UpdateFile(s, t)		# Creates new target file.
							continue
						if s.attributes.user != t.attributes.user and flags & DELTA_FILE_UGM:
							yield UpdateUser(t, s.attributes.user)
						if s.attributes.group != t.attributes.group and flags & DELTA_FILE_UGM:
							yield UpdateGroup(t, s.attributes.group)
						if s.attributes.mode != t.attributes.mode and flags & DELTA_FILE_UGM:
							yield UpdateMode(t, s.attributes.mode)
				elif type(s) == StorageManifest:
					if flags & DELTA_FILE_REMOVE and flags & DELTA_FOLDER_ADD: yield ReplaceWithFolder(s, t)
				else:
					if flags & DELTA_FOLDER_REMOVE and flags & DELTA_FILE_ADD: yield ReplaceWithFile(s, t)
			else:
				# No t - add file/folder
				if isinstance(s, StorageManifest):
					if not target.path.startswith(s.path):
						if flags & DELTA_FOLDER_ADD: yield AddFolder(s, target)
				else:
					if flags & DELTA_FILE_ADD: yield AddFile(s, target)
		else:
			# No s - trim from target
			if isinstance(t, StorageManifest):
				if flags & DELTA_FOLDER_REMOVE: yield RemoveFolder(t)
			else:
				if flags & DELTA_FILE_REMOVE: yield RemoveFile(t)

#
#
def folder_transfer(self, delta, target):
	"""An async routine to copy folders-and-files to target folder."""

	# Interruption happens at the lowest level of transfer activity, i.e before
	# each IO operation (i.e. read or write of blocks). Parent object calls
	# a halt() on this object and the halted flag is checked by every delta
	# opcode (e.g. storage.AddFile) before every IO. If halted it raises the special
	# TransferHalted exception to jump back to this code.
	machine = DeltaMachine(self, target)
	try:
		deltas = len(delta)
		self.console(f'File transfer ({deltas} deltas) to {target}')
		for d in delta:
			d(machine)
			aliases = machine.aliases()
		self.console(f'Move {aliases} aliases to targets')
		machine.rename()
	except TransferHalted:
		return Aborted()
	except OSError as e:
		condition = str(e)
		fault = Faulted(condition)
		return fault
	finally:
		aliases = machine.aliases()
		self.console(f'Clear {aliases} aliases')
		machine.clear()
	return Ack()

bind(folder_transfer)

#
#
class FolderTransfer(Point, StateMachine):
	def __init__(self, delta, target):
		Point.__init__(self)
		StateMachine.__init__(self, INITIAL)
		self.delta = delta
		self.target = target
		self.transfer = None

def FolderTransfer_INITIAL_Start(self, message):
	self.transfer = self.create(folder_transfer, self.delta, self.target)
	return RUNNING

def FolderTransfer_RUNNING_Returned(self, message):
	message = message.message
	self.complete(message)

def FolderTransfer_RUNNING_Stop(self, message):
	halt(self.transfer)
	return HALTED

def FolderTransfer_HALTED_Returned(self, message):
	self.complete(Aborted())

FOLDER_TRANSFER_DISPATCH = {
	INITIAL: (
		(Start,), ()
	),
	RUNNING: (
		(Returned, Stop), ()
	),
	HALTED: (
		(Returned,), ()
	),
}

bind(FolderTransfer, FOLDER_TRANSFER_DISPATCH)



# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Processing of the command line.

Scanning for flags, splitting of flags into name-value pairs and
converting the values according to a relevant schema.
"""
__docformat__ = 'restructuredtext'

__all__ = [
	'process_flags',
	'break_arguments',
	'decode_argument',
	'encode_argument',
	'from_any',
	'extract_arguments',
	'apply_arguments',
	'kv_arguments',
	'remaining_arguments',
	'command_arguments',
	'command_sub_arguments',
	'command_variables',
]

import os
import sys
import json
from .virtual_memory import *
from .convert_signature import *
from .message_memory import *
from .virtual_codec import *
from .json_codec import *
from .noop_codec import *
from .virtual_runtime import *
from .point_runtime import *
from .command_line import *
from .file_object import *


def process_flags(flags):
	'''Scan a raw list of sys args. Divide into words and flags.'''
	word = []
	lf = {}		# Long form.
	sf = {}		# Short.

	for f in flags:
		if f.startswith('--'):
			i = f.find('=')
			if i == -1:
				k = f[2:]
				v = None
			else:
				k = f[2:i]
				v = f[i + 1:]
			lf[k] = v
		elif f.startswith('-'):
			i = f.find('=')
			if i == -1:
				k = f[1:]
				v = None
			else:
				k = f[1:i]
				v = f[i + 1:]
			sf[k] = v
		else:
			word.append(f)

	return word, (lf, sf)

def break_arguments(arguments):
	'''Process a command line. Return executable, words and flags.'''
	executable = os.path.abspath(arguments[0])
	word, flags = process_flags(arguments[1:])
	return executable, word, flags

def includes_pointer(t):
	"""."""
	if isinstance(t, (ArrayOf, VectorOf, DequeOf, SetOf)):
		return includes_pointer(t.element)
	elif isinstance(t, MapOf):
		return includes_pointer(t.value) or includes_pointer(t.key)
	elif isinstance(t, UserDefined):
		x = t.element.__art__
		for k, v in x.schema.items():
			if includes_pointer(v):
				return True
		return False
	elif isinstance(t, PointerTo):
		return True
	elif isinstance(t, Any):
		return True
	return False

# Conversion of data between dicts of json text and a user-defined
# object, i.e. class. The type of each named member controls the
# encode/decode operation.
def w2p(c, w, t):
	return w

# A map of types that are passed as json strings, e.g. times, UUIDs
# and enums.
ENQUOTED = {
	Character: w2p_string,
	Rune: w2p,
	Block: w2p_block,
	String: w2p_string,
	Unicode: w2p,
	Enumeration: w2p_enumeration,
	ClockTime: w2p_clock,
	TimeSpan: w2p_span,
	WorldTime: w2p_world,
	TimeDelta: w2p_delta,
	UUID: w2p_uuid,
	Type: w2p_type,
	Any: w2p_any,
}

def decode_argument(c, s, t):
	'''Convert the json text into a python value based on type. Return decoded value.'''
	q = ENQUOTED.get(type(t), None)
	if q is not None:
		d = q(c, s, t)				# Direct call to the internal conversion.
	elif s is None:
		if isinstance(t, Boolean):	# Empty bool is True.
			return True
		return None
	elif includes_pointer(t):
		# Forced to adopt the full encoding layout.
		d = c.decode(s, t)
	else:
		# Encoding needs to be inserted into the broader
		# packaging of an object.
		j = f'{{"value": {s} }}'
		d = c.decode(j, t)
	return d

def p2w(c, p, t):
	return p

DEQUOTED = {
	#Boolean: lambda c, p, t: 'true' if p else 'false',
	Character: p2w_string,
	Rune: p2w,
	Block: p2w_block,
	String: p2w_string,
	Unicode: p2w,
	Enumeration: p2w_enumeration,
	ClockTime: p2w_clock,
	TimeSpan: p2w_span,
	WorldTime: p2w_world,
	TimeDelta: p2w_delta,
	UUID: p2w_uuid,
	Type: p2w_type,
	Any: p2w_any,
}

def encode_argument(c, p, t):
	'''Convert the python value into JSON text based on type. Return encoded string.'''
	q = DEQUOTED.get(type(t), None)
	if q is not None:
		s = q(c, p, t)
	elif p is None:
		if isinstance(t, Boolean):
			return 'true'
		return 'null'
	elif includes_pointer(t):
		# Forced to adopt the full encoding layout.
		j = CodecJson()
		s = j.encode(p, t)
	else:
		w = c.encode(p, t)
		s = json.dumps(w['value'])
	return s

def first_letters(k):
	'''Generate an acronym for k. Return a string.'''
	s = k.split('_')
	v = [t[0] for t in s if t]	# Omit empty strings, e.g. "from_".
	c = ''.join(v)
	return c

def to_any(v, t):
	if hasattr(v, '__art__') or v is None:
		return v
	return (v, t)

def from_any(v):
	if isinstance(v, tuple) and len(v) == 2 and isinstance(v[1], Portable):
		return v[0]
	return v

def extract_arguments(schema, ls):
	'''Generate a dict of decoded arguments for the given schema and flags. Return a dict and remaining flags.'''
	lf, sf = ls

	# Compile a set of acronyms as aliases for the long form.
	acronym = {}
	for k in schema.keys():
		f = first_letters(k)
		a = acronym.get(f, None)
		if a is None:
			a = []
			acronym[f] = a
		a.append(k)

	# Promote short name to the matching long form.
	def short_to_long(a):
		f = acronym.get(a, None)
		if f is None:
			return None
		if len(f) > 1:
			clashes = ', '.join(f)
			name = f[0].replace('_', '-')
			e1 = f'ambiguous short settings ({clashes})'
			e2 = f'{e1}, use long form --{name}=<value>'
			raise ValueError(e2)
		return f[0]

	# Rather detailed processing of - (dash) and -- (dash-dash)
	# arguments.
	extracted = {}
	lr, sr = {}, {}
	c = CodecJson()

	# Long form --<name>=<value>
	for k, v in lf.items():
		r = k.replace('-', '_')
		t = schema.get(r, None)
		if t is not None:
			try:
				d = decode_argument(c, v, t)	# Matched.
				extracted[r] = to_any(d, t)
			except ValueError as e:
				raise ValueError(f'cannot decode value for "{k}" ({e})')
			except KeyError as e:
				raise ValueError(f'cannot decode value for "{k}" ({e.args[0]} does not exist)')
			except CodecError as e:
				s = str(e)
				s = s.replace('cannot decode', f'cannot decode value for "{k}"')
				raise ValueError(s)
		else:
			lr[k] = v			# Not matched - remainder.

	# Short form -<first-letters>=<value>
	for k, v in sf.items():
		t = short_to_long(k)	# Map to long name.
		if t is None:
			sr[k] = v			# No mapping in schema.
			continue
		r = t.replace('_', '-')
		if t in extracted:
			raise ValueError(f'cannot decode value for "{k}" (use of both long and short forms)')
		x = schema.get(t, None)
		if x is not None:
			try:
				d = decode_argument(c, v, x)
				extracted[t] = to_any(d, x)
			except ValueError as e:
				raise ValueError(f'cannot decode value for "{k}" ({e})')
			except KeyError as e:
				raise ValueError(f'cannot decode value for "{k}" ({e.args[0]} does not exist)')
			except CodecError as e:
				s = str(e)
				s = e.replace('cannot decode', f'cannot decode value for "{k}"')
				raise ValueError(s)
		else:
			sr[k] = v

	return extracted, (lr, sr)

def apply_arguments(target, extracted):
	'''Assign members of the target with named values.'''
	schema = type_schema(target)

	for k, v in extracted.items():
		if k in schema:
			d = from_any(v)
			setattr(target, k, d)

def kv_arguments(extracted):
	kv = {k: from_any(v) for k, v in extracted.items()}
	return kv

def remaining_arguments(remainder):
	argument = []
	for k, v in remainder[0].items():
		if v:
			argument.append(f'--{k}={v}')
		else:
			argument.append(f'--{k}')
	for k, v in remainder[1].items():
		if v:
			argument.append(f'-{k}={v}')
		else:
			argument.append(f'-{k}')
	return argument

def command_arguments(object_type, override_arguments=None):
	'''Process sys.argv according to schemas in CommandLine and object_type. Return decoded info.'''
	process_schema = type_schema(CL)
	object_schema = type_schema(object_type)
	c = set(process_schema.keys()) & set(object_schema.keys())
	if len(c) > 0:
		j = ', '.join(c)
		raise ValueError(f'collision in settings names - {j}')
	if override_arguments is None:
		arguments = sys.argv
	else:
		arguments = override_arguments
	executable, word, flags = break_arguments(arguments)
	extracted, remainder = extract_arguments(process_schema, flags)
	apply_arguments(CL, extracted)
	extracted, remainder = extract_arguments(object_schema, remainder)
	if remainder[0] or remainder[1]:
		k = [k for k in remainder[0].keys()]
		k.extend([k for k in remainder[1].keys()])
		r = ', '.join(k)
		raise ValueError(f'unknown arguments ({r})')
	return executable, extracted, word

#
def break_table(object_table):
	table = {t.__name__.rstrip('_'): (t, type_schema(t)) for t in object_table}
	return table

def command_sub_arguments(object_type, object_table, override_arguments=None, strict=True):
	'''Process sys.argv according to schemas in CommandLine, object_type and sub-commands. Return decoded info.'''
	process_schema = type_schema(CL)
	object_schema = type_schema(object_type)
	s1 = set(process_schema.keys())
	s2 = set(object_schema.keys())
	c = s1 & s2
	if len(c) > 0:
		j = ', '.join(c)
		raise ValueError(f'collision in settings names - {j}')
	if override_arguments is None:
		arguments = sys.argv
	else:
		arguments = override_arguments
	executable, word, flags = break_arguments(arguments)
	extracted, remainder = extract_arguments(process_schema, flags)
	apply_arguments(CL, extracted)
	objected, remainder = extract_arguments(object_schema, remainder)

	# Optional appearance of a sub-command in the arguments.
	if len(word) > 0:
		sub_command = word[0]

		table = break_table(object_table)
		t = table.get(sub_command, None)
		if t is None:
			raise ValueError(f'unknown sub-command "{sub_command}"')
		sub_schema = t[1]
		if sub_schema is None:
			raise ValueError(f'cannot use sub-command "{sub_command}", not registered?')
		s3 = set(sub_schema.keys())
		s4 = s1 | s2
		c = s3 & s4
		if len(c) > 0:
			j = ', '.join(c)
			raise ValueError(f'collision in settings names - {j}')
		extracted, remainder = extract_arguments(sub_schema, remainder)
		jump = t[0]
		kv = kv_arguments(extracted)
		word[0] = (sub_command, jump, kv, remainder)

	#else:
	#	sub = None		# Non sub command.

	if strict and (remainder[0] or remainder[1]):
		k = [k for k in remainder[0].keys()]
		k.extend([k for k in remainder[1].keys()])
		r = ', '.join(k)
		raise ValueError(f'unknown arguments ({r})')
	return executable, objected, word

def command_variables(factory_variables, environment_variables=None):
	'''Scan environment for values matching the given schema. Update the given values.'''
	if factory_variables is None:
		return

	environ = environment_variables or os.environ
	environment_variables = {k[5:].replace('-', '_'): v for k, v in environ.items()
		if k.startswith('LC_V_')}

	schema = type_schema(factory_variables)
	c = CodecJson()
	for k, t in schema.items():
		K = k.upper()
		e = environment_variables.get(K, None)
		if e is None:
			continue
		try:
			v = decode_argument(c, e, t)
			setattr(factory_variables, k, v)
		except ValueError as e:
			raise ValueError(f'cannot decode value for "{k}" ({e})')
		except CodecError as e:
			s = str(e)
			raise ValueError(f'cannot process variable "{K}" ({k}) - {s}')



# test_storage.py
import kipjak as kj

#
#
def storage(self, disk_path: str=None) -> kj.StorageManifest:
	'''.'''
	if disk_path is None:
		return kj.Faulted('no storage specified.')

	m = kj.storage_manifest(disk_path)
	return m[0]

kj.bind(storage)	# Register with the framework.



if __name__ == '__main__':	# Process entry-point.
	kj.create(storage)



# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implementation of a passthru codec.

"""

__docformat__ = 'restructuredtext'

from .virtual_memory import *
from .convert_memory import *
from .virtual_runtime import *
from .message_memory import *
from .virtual_codec import *


__all__ = [
	'word_to_noop',
	'noop_to_word',
	'CodecNoop'
]

def word_to_noop(c, w):
	return w

def noop_to_word(c, s):
	return s

class CodecNoop(Codec):
	EXTENSION = 'noop'

	def __init__(self, return_proxy=None, local_termination=None, pretty_format=False, decorate_names=True):
		Codec.__init__(self,
			CodecNoop.EXTENSION,
			word_to_noop,
			noop_to_word,
			return_proxy, local_termination, pretty_format, decorate_names)



# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Machines send messages and dispatch received machines to functions.

Purest async objects. Capable of sharing a thread.
"""
__docformat__ = 'restructuredtext'

import re

from .virtual_memory import *
from .message_memory import *
from .convert_signature import *
from .convert_type import *
from .virtual_runtime import *
from .point_runtime import *
from .virtual_point import *

__all__ = [
	'Stateless',
	'StateMachine',
	'DEFAULT',
]

class DEFAULT: pass

# Find the state and message embedded within a function name.
state_message = re.compile('(?P<state>[A-Z][A-Z0-9]*(_[A-Z0-9]+)*)_(?P<message>[A-Z][A-Za-z0-9]*)')

unknown = portable_to_signature(UserDefined(Unknown))

class Stateless(Machine):
	"""Base for simple machines that maintain no formal state.

	Messages are received by an assigned thread and dispatched to
	handlers according to the type of the received message.
	"""
	def __init__(self):
		Machine.__init__(self)

	def transition(self, message):
		art = self.__art__
		shift, messaging = art.value
		m, p, a = un_cast(message)
		s = portable_to_signature(p)
		f = shift.get(s, None)			# Explicit match.
		if f:
			return m, p, a, f

		if a:
			for c, f in messaging.items():
				if isinstance(m, c):		# Base-derived match.
					return m, p, a, f

		f = shift.get(unknown, None)	# Catch-all.
		return m, p, a, f

	def received(self, queue, message, return_address):
		"""Dispatch message to the appropriate handler.

		:parm queue: instance of a Queue-based async object
		:type queue: a Queue-based async class
		:parm message: the received massage
		:type message: instance of a registered class
		:parm return_address: origin of the message
		:type return_address: object address
		:rtype: none
		"""
		art = self.__art__
		m, p, a, f = self.transition(message)
		r = return_address[-1]
		if f is None:
			if art.message_trail and (not a or a.message_trail):
				t = portable_to_tag(p)
				if isinstance(m, Faulted):
					e = str(m)
					self.log(USER_TAG.RECEIVED, f'Dropped {t} from <{r:08x}>, {e}')
				else:
					self.log(USER_TAG.RECEIVED, f'Dropped {t} from <{r:08x}>')
			return

		if art.message_trail and (not a or a.message_trail):
			t = portable_to_tag(p)
			if isinstance(m, Faulted):
				e = str(m)
				self.log(USER_TAG.RECEIVED, f'Received {t} from <{r:08x}> {e}')
			else:
				self.log(USER_TAG.RECEIVED, f'Received {t} from <{r:08x}>')

		self.received_type = p
		f(self, m)

class StateMachine(Machine):
	"""Base for machines that maintain a formal state.

	Messages are received by an assigned thread and dispatched to
	handlers according to the current state and the type of the
	received message.

	Every handler must return the next state. In those cases where
	the state remains unchanged, return ``self.current_state``. Termination
	of a machine is by a call to :meth:`~.Point.complete`.

	:param initial: Start state for all instances of derived class
	:type initial: class
	"""
	def __init__(self, initial):
		Machine.__init__(self)
		self.current_state = initial

	def transition(self, state, message):
		art = self.__art__
		shift, messaging = art.value
		m, p, a = un_cast(message)
		s = portable_to_signature(p)
		shifted = shift.get(state, None)
		if shifted is None:
			raise ValueError(f'machine "{art.path}" shifted to nowhere')

		f = shifted.get(s, None)				# Explicit match.
		if f:
			return m, p, a, f

		if a:
			messaged = messaging.get(state, None)
			if messaged:
				for c, f in messaged.items():
					if isinstance(m, c):			# Base-derived match.
						return m, p, a, f

		f = shifted.get(unknown, None)			# Catch-all.
		return m, p, a, f

	def received(self, queue, message, return_address):
		"""Dispatch message to the appropriate handler.

		:parm queue: instance of a Queue-based async object
		:type queue: a Queue-based async class
		:parm message: the received massage
		:type message: instance of a registered class
		:parm return_address: origin of the message
		:type return_address: object address
		:rtype: none
		"""
		art = self.__art__
		m, p, a, f = self.transition(self.current_state, message)
		r = return_address[-1]
		if f is None:
			if art.message_trail and (not a or a.message_trail):
				t = portable_to_tag(p)
				if isinstance(message, Faulted):
					e = str(message)
					self.log(USER_TAG.RECEIVED, f'Dropped {t} from <{r:08x}>, {e}')
				else:
					self.log(USER_TAG.RECEIVED, f'Dropped {t} from <{r:08x}>')
			return

		if art.message_trail and (not a or a.message_trail):
			t = portable_to_tag(p)
			if isinstance(message, Faulted):
				e = str(message)
				self.log(USER_TAG.RECEIVED, f'Received {t} from <{r:08x}> {e}')
			else:
				self.log(USER_TAG.RECEIVED, f'Received {t} from <{r:08x}>')

		self.received_type = p
		self.current_state = f(self, m)



# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Logging methods available within the async runtime.

"""
__docformat__ = 'restructuredtext'

import os
import sys
import time

from .virtual_runtime import *

__all__ = [
	'PID',
	'log_to_nowhere',
	'log_to_stderr',
	'select_logs',
]

# Some essential logging options.
PID = os.getpid()

def log_to_nowhere(log):
	pass

def log_to_stderr(log):
	second = time.strftime('%Y-%m-%dT%H:%M:%S', time.gmtime(log.stamp))
	fraction = '%.3f' % (log.stamp,)
	fraction = fraction[-3:]
	mark = '%s.%s' % (second, fraction)
	name = log.name.split('.')[-1]
	state = log.state
	if state is None:
		p = '[%08d] %s %s <%08x>%s - %s\n' % (PID, mark, log.tag.value, log.address[-1], name, log.text)
	else:
		p = '[%08d] %s %s <%08x>%s[%s] - %s\n' % (PID, mark, log.tag.value, log.address[-1], name, state, log.text)
	sys.stderr.write(p)
	sys.stderr.flush()

def select_logs(tag):
	def log_by_number(log):
		t = tag_to_log(log.tag)
		if tag.value > t.value:
			return
		log_to_stderr(log)
	return log_by_number



#
#
import kipjak as kj

# MESSAGES ------------------------------------------------------------------------------
class Cook(object):
	def __init__(self, a: int=None, b: dict[int, str]=None, c=None):	# Cant hint here (-> int). Would annoy checkers.
		self.a = a
		self.b = b
		self.c = c

class Radiating(object):
	pass

kj.bind_message(Cook, c=kj.PointerTo(kj.Boolean))
kj.bind_message(Radiating)

# 1. List of message contents and their types (as hints)
# 2. Still must be default constructible (flexible matching policy on decoding)
# 3. Types provided on bind when hints fall short.
# 4. Hints use Python syntax, bind types expect Portable (unfixed) but accept hint or class as well.

# POINT AS FUNCTION ---------------------------------------------------------------------
BR_1 = 1	#kj.branch_table(int, dict[int, str])
JP_1 = 2	#kj.jump_table(a=int, b=dict[int, str])

def main(self, a: int=None, b: dict[int, str]=None, c=None) -> int:
	b = b or {}

	# self.select(h, h, h, ...) -> matching message
	# self.branch(table) -> matching index and message
	# self.jump -> matching index and message after call to jump routine

	isinstance(a, int)
	isinstance(b, dict)

	i, m = self.branch(BR_1)

	def a(self, i):
		return i + 1

	i, r = self.jump(JP_1, a=a)

	return 0

kj.bind_routine(main, c=kj.TimeSpan())

#
#
def poly(self, a: int=None, b: dict[int, str]=None, c=None) -> kj.Any:
	b = b or {}

	isinstance(a, int)
	isinstance(b, dict)

	m, i = self.branch(BR_1)
	if i == 0:
		return (0, None)

	return ({'a':1}, None)

kj.bind_routine(main, c=kj.TimeSpan())



# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Format startup and shutdown of the async runtime.

Ensure that the support for async operation is in place when the process
needs it. Ensure that support is cleared out during process termination.
"""
__docformat__ = 'restructuredtext'

import os
import threading
import atexit
from .object_space import *
from .convert_type import *
from .virtual_runtime import *
from .virtual_point import *
from .point_runtime import *
from .general_purpose import *
from .object_logs import *
from .log_agent import *
from .countdown_timer import *
from .bind_type import *

__all__ = [
	'PB',
	'DEFAULT_HOME',
	'AddOn',
	'QuietChannel',
	'start_up',
	'tear_down',
	'open_channel',
	'drop_channel',
	'channel',
]

PB = Gas(
	tear_down_atexit=True,
	root=None,					# The top of the hierarchy.
	thread_dispatch={},
	add_ons=[],
	output_value=None,
	exit_status=None,
)

DEFAULT_HOME = '.kipjak'

# Register create and stop functions for some
# point-based singleton.
class AddOn(object):
	def __init__(self, create, stop):
		cs = (create, stop)
		PB.add_ons.append(cs)

# A silent intermediary between the worlds of sync and
# async.
class QuietChannel(Channel):
	def __init__(self):
		Channel.__init__(self)

bind_point(QuietChannel, lifecycle=False, message_trail=False, execution_trace=False, user_logs=USER_LOG.NONE)

# A non-logging channel.
root_lock = threading.RLock()

def start_up(logs=log_to_nowhere):
	"""Start the async runtime. Return the root object.

	This is the function that actually creates the threads and objects
	that are needed to support Point operations, e.g. timers. It also
	arranges for automated cleanup using atexit.

	It first checks to see if the runtime is already up, in a thread-safe
	way. Only the first call has any effect.

	:param logs: an object expecting to receive log objects
	:type logs: a callable object
	:rtype: channel
	"""
	global root_lock
	try:
		root_lock.acquire()
		root = PB.root
		if root is None:
			if PB.tear_down_atexit:
				atexit.register(tear_down)

			nowhere = Point()

			root = nowhere.create(QuietChannel)

			PB.root = root
			VP.log_address = root.create(LogAgent, logs)
			VP.timer_address = root.create(CountdownTimer)
			#VP.test_address = root.create(TestRecord)
			VP.circuit_address = root.create(timer_circuit, VP.timer_address)
			bg = root.create(object_dispatch)
			set_queue(None, bg)
			for k, s in VP.thread_classes.items():
				t = root.create(object_dispatch)
				for c in s:
					set_queue(c, t)
				PB.thread_dispatch[k] = t
			for cs in PB.add_ons:
				cs[0](root)
	finally:
		root_lock.release()
	return root

def tear_down():
	"""End the async runtime. Returns nothing.

	:rtype: None
	"""
	global root_lock
	try:
		root_lock.acquire()
		root = PB.root
		if root:
			s = Stop()
			for cs in reversed(PB.add_ons):
				cs[1](root)
			for _, t in PB.thread_dispatch.items():
				root.send(s, t)
				root.select(Returned)
			bg = get_queue_address(None)
			root.send(s, bg)
			root.select(Returned)
			halt(VP.circuit_address)
			root.select(Returned)
			root.send(s, VP.timer_address)
			root.select(Returned)
			root.send(s, VP.log_address)
			root.select(Returned)
			drop_channel(root)
			PB.root = None
	finally:
		root_lock.release()

	exit_status = PB.exit_status
	if exit_status is not None:
		os._exit(exit_status)

# Access to async from sync section of an
# application.
def open_channel():
	"""Start the runtime for a non-standard executable. Return a unique async object.

	Create a new async object for the purposes of initiating
	async activity, typically from within a non-async section of
	code. Registers a cleanup function with the process, to execute
	on process termination.

	:rtype: Point-based object
	"""
	root = start_up()
	channel = root.create(Channel)
	return channel

def drop_channel(c: Point):
	"""End the runtime for a non-standard executable. Return nothing.

	Tear down the runtime created by ``open_channel()``, i.e. ``boot_up()``.

	:param c: a channel returned by open_channel()
	:rtype: none
	"""
	if c.__art__.lifecycle:
		c.log(USER_TAG.DESTROYED, 'Destroyed')
	destroy_an_object(c.object_address)

#
#
class channel(object):
	"""A context to automate the opening and closing of a channel.

	Typically used in a traditional, sync application to access the async
	features of kipjak. May be used anywhere within an application. Each
	instance creates a unique channel. The parameter provides for control
	over the fate of logs. This only has an effect on the first use of
	the class.

	:param logs: an object expecting to receive log objects
	:type logs: a callable object
	"""
	def __init__(self):
		self.channel = open_channel()

	def __enter__(self):
		return self.channel

	def __exit__(self, exc_type=None, exc_value=None, traceback=None):
		drop_channel(self.channel)

		# Handle exceptions that may have come up during execution, by
		# default any exceptions are raised to the user.
		if exc_type is not None:
			return False
		return True



# object_startup_test.py
import kipjak as kj
import test_main

class Main(kj.Point, kj.Stateless):
	def __init__(self):
		kj.Point.__init__(self)
		kj.Stateless.__init__(self)
		self.temperature = 0
		self.timeout = 0

def Main_Start(self, message):
	t = kj.text_to_world('1963-03-26T02:24')
	a = self.create(kj.ProcessObject, test_main.main, b=32, c=99, t=t)

	def test_main_complete(self, value, _):
		self.complete(value)

	self.on_return(a, test_main_complete)

def Main_Returned(self, message):
	d = self.debrief()
	if isinstance(d, kj.OnReturned):
		d(self, message)

def Main_Stop(self, message):
	self.abort()

kj.bind(Main, dispatch=(kj.Start, kj.Returned, kj.Stop), return_type=kj.Any())

if __name__ == '__main__':
	kj.create(Main)



# test_subscriber.py
import kipjak as kj

from test_api import *

# A bare-bones implementation of a client subscribing to a network service.
DEFAULT_SEARCH = 'acme'

class Subscriber(kj.Point, kj.Stateless):
	def __init__(self, search: str=None, seconds: float=None, enduring: bool=False):
		kj.Point.__init__(self)
		kj.Stateless.__init__(self)
		self.search = search or DEFAULT_SEARCH
		self.seconds = seconds
		self.enduring = enduring

def Subscriber_Start(self, message):
	kj.subscribe(self, self.search)		# Connect this object with the named object.
	if self.seconds is not None:
		self.start(kj.T1, 5.0)			# Expect resolution with a few seconds.

def Subscriber_Available(self, message):
	self.send(Xy(x=2, y=2), self.return_address)

def Subscriber_list_list_float(self, message):
	if self.enduring:
		return
	self.complete(message)

def Subscriber_T1(self, message):
	self.complete(kj.TimedOut(message))

def Subscriber_Faulted(self, message):		# All faults routed here including
	self.complete(message)					# failure of subscribe().

def Subscriber_Stop(self, message):
	self.complete(kj.Aborted())			# Leave all the housekeeping to the framework.

kj.bind(Subscriber, (kj.Start, kj.Available, table_type, kj.T1, kj.Faulted, kj.Stop), return_type=table_type)


if __name__ == '__main__':		# Process entry-point.
	kj.create(Subscriber)



# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
""" Dedicated collector for objects.

Objects register themselves with a collector during their start process
and deregister during teardown. Addresses still in the collector at
termination are signaled directly - the collector waits for all objects
to clear themselves.

Originally developed for garbage collection of ProcessObjects.
"""
__docformat__ = 'restructuredtext'

from queue import Queue, Full, Empty
from collections import deque

from .virtual_memory import *
from .message_memory import *
from .virtual_runtime import *
from .point_runtime import *
from .virtual_point import *
from .point_machine import *
from .bind_type import *

__all__ = [
	'AddObject',
	'RemoveObject',
	'ObjectCollector',
]

# Register and deregister.
class AddObject(object):
	def __init__(self, address: Address=None):
		self.address = address or NO_SUCH_ADDRESS

class RemoveObject(object):
	def __init__(self, address: Address=None):
		self.address = address or NO_SUCH_ADDRESS

bind(AddObject)
bind(RemoveObject)

class INITIAL: pass
class READY: pass
class CLEARING: pass

class ObjectCollector(Threaded, StateMachine):
	def __init__(self, grace: float=10.0):
		Threaded.__init__(self)
		StateMachine.__init__(self, INITIAL)
		self.grace = grace
		self.collected = set()

def ObjectCollector_INITIAL_Start(self, message):
	return READY

def ObjectCollector_READY_AddObject(self, message):
	self.collected.add(message.address)
	return READY

def ObjectCollector_READY_RemoveObject(self, message):
	self.collected.discard(message.address)
	return READY

def ObjectCollector_READY_Stop(self, message):
	if len(self.collected) < 1:
		self.complete()
	for c in self.collected:
		self.send(message, c)
	self.start(T1, self.grace)
	return CLEARING

def ObjectCollector_CLEARING_AddObject(self, message):
	self.collected.add(message.address)
	self.send(Stop(), message.address)
	return CLEARING

def ObjectCollector_CLEARING_RemoveObject(self, message):
	self.collected.discard(message.address)
	if len(self.collected) < 1:
		self.complete()
	return CLEARING

def ObjectCollector_CLEARING_T1(self, message):
	self.complete()

OBJECT_COLLECTOR_DISPATCH = {
	INITIAL: (
		(Start,),
		()
	),
	READY: (
		(AddObject, RemoveObject, Stop,),
		()
	),
	CLEARING: (
		(AddObject, RemoveObject, T1,),
		()
	),
}

bind(ObjectCollector, OBJECT_COLLECTOR_DISPATCH)



# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Temporary, thread-safe buffering of sent messages and how to get messages with reordering.

A collection of classes to be inherited by different points. Most notably
synchronous channels, and the 2 different machine types.
"""
__docformat__ = 'restructuredtext'

from queue import Queue, Full, Empty
from collections import deque

from .virtual_memory import *
from .message_memory import *
from .virtual_runtime import *
from .point_runtime import *

__all__ = [
	'PEAK_BEFORE_DROPPED',
	'Pump',
]

PEAK_BEFORE_DROPPED = 16384
GRACE_PERIOD = 5

# The buffering between senders and receivers. Firstly this is a
# wrapper around the system queue. Then there are several flavours
# of access to that wrapper. One style about sync access and the
# other about message processing for machines. Both of them implementing
# the save-replay model from SDL.
class Pump(object):
	"""Base for any object intended to operate as a message queue.

	:param blocking: behaviour on queue full
	:type blocking: bool
	:param maximum_size: number of message to hold
	:type maximum_size: int
	"""

	def __init__(self, blocking=False, maximum_size=PEAK_BEFORE_DROPPED):
		"""Construct an instance of pump."""
		self.blocking = blocking
		self.message_queue = Queue(maxsize=maximum_size)
		self.thread_function = None
		self.assigned_thread = None

	def put(self, mtr):
		"""Append the [message, to, return] triple to the queue."""
		try:
			self.message_queue.put(mtr, self.blocking)
		except Full:
			# Silently FOTF.
			pass

	def get(self):
		"""Return the pending [message, to, return] triplet or block."""
		mtr = self.message_queue.get()
		return mtr



# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""The directory built in to every process. Default is inactive.

.
"""
__docformat__ = 'restructuredtext'

from enum import Enum
from collections import deque

from .general_purpose import *
from .ip_networking import *
from .command_line import *
from .point_runtime import *
from .virtual_point import *
from .object_runtime import *
from .object_directory import *

__all__ = [
	'PD',
	'publish',
	'subscribe',
	'clear_published',
	'clear_subscribed',
]

PD = Gas(directory=None)

'''
ifs = netifaces.interfaces()
for i in ifs:
	a = netifaces.ifaddresses(i)
	if netifaces.AF_INET not in a:
		continue
	print(a[netifaces.AF_INET][0]['addr'])

'''

# Managed creation of the builtin directory.
def create_directory(root):
	directory_scope = CL.directory_scope or ScopeOfDirectory.PROCESS
	connect_to_directory = CL.connect_to_directory or HostPort()
	accept_directories_at = CL.accept_directories_at or HostPort()
	PD.directory = root.create(ObjectDirectory, directory_scope=directory_scope,
		connect_to_directory=connect_to_directory,
		accept_directories_at=accept_directories_at)

def stop_directory(root):
	root.send(Stop(), PD.directory)
	root.select(Returned)

AddOn(create_directory, stop_directory)

#
def publish(self: Point, name: str, scope: ScopeOfDirectory=ScopeOfDirectory.HOST, encrypted: bool=False):
	"""
	Establish a service presence under the specified name.

	:param self: asynchronous identity
	:param name: name to be used as alias for the given object
	:param scope: scope in which the name is available
	:param encrypted: enable encryption of subscriber sessions
	"""
	p = PublishAs(name=name, scope=scope, publisher_address=self.object_address, encrypted=encrypted)
	self.send(p, PD.directory)

def subscribe(self: Point, search: str, scope: ScopeOfDirectory=ScopeOfDirectory.HOST):
	"""
	Establish a lookup for services matching the specified pattern.

	:param self: asynchronous identity
	:param search: pattern to be used for matching with services
	:param scope: scope in which to search
	"""
	p = SubscribeTo(search=search, scope=scope, subscriber_address=self.object_address)
	self.send(p, PD.directory)

def clear_published(self: Point, published: Published, note: str=None):
	"""
	Remove all trace of the previously published service.

	:param self: asynchronous identity
	:param published: message confirming a :func:`~.publish`
	:param note: short description added to logs
	"""
	p = ClearPublished(name=published.name, scope=published.scope, published_id=published.published_id, note=note)
	self.send(p, PD.directory)

def clear_subscribed(self: Point, subscribed: Subscribed, note: str=None):
	"""
	Remove all trace of the previously registered search.

	:param self: asynchronous identity
	:param published: message confirming a :func:`~.subscribe`
	:param note: short description added to logs
	"""
	s = ClearSubscribed(search=subscribed.search, scope=subscribed.scope, subscribed_id=subscribed.subscribed_id, note=note)
	self.send(s, PD.directory)



# test_main_args.py
import datetime
import uuid
import kipjak as kj

from test_person import *

class Main(kj.Point, kj.Stateless):
	def __init__(self, table: dict[str,list[Person]]=None,
		count: int=10, ratio: float=0.5,
		who: Person=None, when: datetime.datetime=None,
		unique_id: uuid.UUID=None):
		kj.Point.__init__(self)
		kj.Stateless.__init__(self)
		self.table = table or dict(recent=[Person('Felicity'), Person('Frederic')])
		self.who = who or Person('Wilfred')
		self.when = when or kj.world_now()
		self.unique_id = unique_id or uuid.uuid4()
		self.count = count
		self.ratio = ratio

def Main_Start(self, message):
	# Name of every person.
	person = []
	for v in self.table.values():
		for p in v:
			person.append(p.given_name)

	# Log values.
	csv = ','.join(person)
	self.console(table=csv)
	self.console(count=self.count, ratio=self.ratio)
	self.console(who=self.who.given_name, when=self.when)
	self.console(unique_id=self.unique_id)

	self.complete()

kj.bind(Main, dispatch=(kj.Start,))

if __name__ == '__main__':
	kj.create(Main)



# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
__docformat__ = 'restructuredtext'

import sys
import types
import typing

from .virtual_memory import *
from .message_memory import *
from .convert_type import *
from .convert_signature import *
from .virtual_runtime import *
from .virtual_point import *
from .point_machine import *

__all__ = [
	'PointRuntime',
	'bind_routine',
	'bind_point',
	'bind_stateless',
	'bind_statemachine',
	'bind',
]

#
class PointRuntime(Runtime):
	"""
	Settings to control logging and other behaviour, for a Point.

	:param name: the name of the class being registered
	:param module: the name of the module the class is located in
	:param return_type: hint/portable describing the return type
	:type return_type: :ref:`tip<type-reference>`
	:param entry_point: enable library loading with list of expected messages
	:param flags: named values passed on
	"""

	def __init__(self,
			name: str, module: str, return_type=None, entry_point: list=None,
			**flags):
		super().__init__(name, module, **flags)
		self.return_type = return_type
		self.entry_point = entry_point
		self.value = None

#
def bind_routine(routine, return_type=None, entry_point: list=None,
		lifecycle: bool=True, message_trail: bool=True, execution_trace: bool=True,
		user_logs: USER_LOG=USER_LOG.DEBUG, **explicit_schema):
	"""
	Set the type information and runtime controls for the function.

	Values assigned in this function affect the behaviour for all instances of
	the given type.

	:param routine: function to be registered as a routine
	:type routine: :ref:`object type<kj-object-type>`
	:param return_type: type expression for the return value
	:type return_type: :ref:`tip<type-reference>`
	:param entry_point: enable library loading with list of expected messages
	:param lifecycle: enable log at creation, ending...
	:param message_trail: enable log when message is sent
	:param execution_trace: enable log when message is received
	:param user_logs: the logging level for this message type
	"""
	rt = PointRuntime(routine.__name__, routine.__module__,
		lifecycle=lifecycle,
		message_trail=message_trail,
		execution_trace=execution_trace,
		user_logs=user_logs)

	setattr(routine, '__art__', rt)

	# Replace with identity object, installing as required.
	explicit_schema = {k: install_type(v) for k, v in explicit_schema.items()}

	hints = typing.get_type_hints(routine)
	routine_hints, routine_return = install_hints(hints)

	if return_type:
		routine_return = install_type(return_type)

	r = {}
	for k, a in explicit_schema.items():
		routine_hints[k] = a	# Add or override existing.

	if entry_point is not None:
		entry_point = [install_type(a) for a in entry_point]

	rt.schema = routine_hints
	rt.return_type = routine_return
	rt.entry_point = entry_point

	install_portable(UserDefined(routine))


def bind_point(point: Point, return_type=None, entry_point: list=None, thread: str=None,
		lifecycle: bool=True, message_trail: bool=True, execution_trace: bool=True,
		user_logs: USER_LOG=USER_LOG.DEBUG, **explicit_schema):
	"""
	Set the type information and runtime controls for the asynchronous object.

	Values assigned in this function affect the behaviour for all instances of
	the given type.

	:param point: instance of an asynchronous object
	:param return_type: type expression for the return value
	:type return_type: :ref:`tip<type-reference>`
	:param lifecycle: enable log when object is created or destroyed
	:param message_trail: enable log when message is sent
	:param execution_trace: enable log when message is received
	:param user_logs: the logging level for this object type
	"""
	rt = PointRuntime(point.__name__, point.__module__,
		lifecycle=lifecycle,
		message_trail=message_trail,
		execution_trace=execution_trace,
		user_logs=user_logs)

	setattr(point, '__art__', rt)

	explicit_schema = {k: install_type(v) for k, v in explicit_schema.items()}

	hints = typing.get_type_hints(point.__init__)
	point_hints, _ = install_hints(hints)

	if return_type:
		return_type = install_type(return_type)

	r = {}
	for k, a in explicit_schema.items():
		point_hints[k] = a	# Add or override existing.

	if entry_point is not None:
		entry_point = [install_hint(a) for a in entry_point]

	rt.schema = point_hints
	rt.return_type = return_type
	rt.entry_point = entry_point

	if thread:
		try:
			q = VP.thread_classes[thread]
		except KeyError:
			q = set()
			VP.thread_classes[thread] = q
		q.add(point)

#
def message_handler(name):
	# Cornered into unusual iteration by test framework.
	# Collection of tests fails with "dict changed its size".
	for k in list(sys.modules):
		v = sys.modules[k]
		if isinstance(v, types.ModuleType):
			try:
				f = v.__dict__[name]
				if isinstance(f, types.FunctionType):
					return f
			except KeyError:
				pass
	return None

def statemachine_save(self, message):
	self.save(message)
	return self.current_state

def unfold(folded):
	for f in folded:
		if isinstance(f, (tuple, list)):
			yield from unfold(f)
		else:
			yield f

def bind_stateless(machine: Stateless, dispatch: tuple, return_type=None, entry_point: list=None, **explicit_schema):
	"""
	Set the type information and runtime controls for the non-FSM machine.

	Values assigned in this function affect the behaviour for all instances of
	the given type.

	:param machine: class to be registered as a machine
	:param dispatch: list of expected messages
	:param return_type: type expression for the return value
	:type return_type: :ref:`tip<type-reference>`
	:param entry_point: enable library loading with list of expected messages
	"""
	bind_point(machine, return_type=return_type, entry_point=entry_point, **explicit_schema)
	if dispatch is None:
		return

	shift = {}
	messaging = {}
	for s in unfold(dispatch):
		p = install_type(s)
		x = portable_to_signature(p)
		d = isinstance(p, UserDefined)
		if d:
			tag = p.element.__name__
		else:
			tag = portable_to_tag(p)
		name = f'{machine.__name__}_{tag}'

		f = message_handler(name)
		if f is None:
			raise PointConstructionError(f'function "{name}" not found ({machine.__art__.path})')

		shift[x] = f
		if d and s is not Unknown:
			messaging[p.element] = f

	machine.__art__.value = (shift, messaging)

def bind_statemachine(machine: StateMachine, dispatch: dict, return_type=None, entry_point: list=None, **explicit_schema):
	"""
	Set the type information and runtime controls for the FSM machine.

	Values assigned in this function affect the behaviour for all instances of
	the given type.

	The dispatch is a description of states, expected
	messages and messages that deserve saving::

		dispatch = {
			STATE_1: (Start, ()),
			STATE_2: ((Job, Pause, UnPause, Stop), (Check,)),
			STATE_3: ((Stop, ()),
		}

	Consider ``STATE_2``; The machine will accept 4 messages and
	will save an additional message, ``Check``.

	:param machine: class to be registered as a machine
	:param dispatch: table of current state and expected messages
	:param return_type: type expression for the return value
	:type return_type: :ref:`tip<type-reference>`
	:param entry_point: enable library loading with list of expected messages
	"""
	bind_point(machine, return_type=return_type, entry_point=entry_point, **explicit_schema)
	if dispatch is None:
		return
	shift = {}
	messaging = {}
	for state, v in dispatch.items():
		if not isinstance(v, tuple) or len(v) != 2:
			raise PointConstructionError(f'FSM {machine.__name__}[{state.__name__}] dispatch is not a tuple or is not length 2')
		matching, saving = v

		if not isinstance(matching, tuple):
			raise PointConstructionError(f'FSM {machine.__name__}[{state.__name__}] (matching) is not a tuple')
		if not isinstance(saving, tuple):
			raise PointConstructionError(f'FSM {machine.__name__}[{state.__name__}] (saving) is not a tuple')

		for m in matching:
			p = install_type(m)
			x = portable_to_signature(p)
			d = isinstance(p, UserDefined)
			e = isinstance(p, Enumeration)
			if d or e:
				tag = p.element.__name__
			else:
				tag = portable_to_tag(p)
			name = '%s_%s_%s' % (machine.__name__, state.__name__, tag)
			f = message_handler(name)
			if f is None:
				raise PointConstructionError(f'function "{name}" not found ({machine.__art__.path})')

			r = shift.get(state, None)
			if r is None:
				r = {}
				shift[state] = r
			r[x] = f

			if d:
				r = messaging.get(state, None)
				if r is None:
					r = {}
					messaging[state] = r
				r[p.element] = f

		for s in saving:
			p = install_type(s)
			x = portable_to_signature(p)
			r = shift.get(state, None)
			if r is None:
				r = {}
				shift[state] = r
			if x in r:
				raise PointConstructionError(f'FSM {machine.__name__}[{state.__name__}] has "{m.__name__}" in both matching and saving')
			r[x] = statemachine_save

			if isinstance(p, UserDefined):
				r = messaging.get(state, None)
				if r is None:
					r = {}
					messaging[state] = r
				r[p.element] = f

	machine.__art__.value = (shift, messaging)

def bind(object_type, *args, **kw_args):
	"""
	Forwards all arguments on to a custom bind function
	according to the type of the first argument.

	- :func:`~.bind_message`
	- :func:`~.bind_routine`
	- :func:`~.bind_stateless`
	- :func:`~.bind_statemachine`

	:param object_type: type of async entity
	:type object_type: :ref:`object type<kj-object-type>`
	:param args: arguments passed to special bind function
	:type args: tuple
	:param kw_args: named arguments passed to special bind function
	:type kw_args: dict
	"""
	# Damn line length constraint.
	if isinstance(object_type, types.FunctionType):
		bind_routine(object_type, *args, **kw_args)
	elif issubclass(object_type, Point):
		if issubclass(object_type, Machine):
			if issubclass(object_type, Stateless):
				bind_stateless(object_type, *args, **kw_args)
			elif issubclass(object_type, StateMachine):
				bind_statemachine(object_type, *args, **kw_args)
			else:
				pce = f'cannot bind {object_type} - unknown machine type'
				raise PointConstructionError(pce)
			install_portable(UserDefined(object_type))
		else:
			bind_point(object_type, *args, **kw_args)
	elif not is_message_class(object_type):
		bind_message(object_type, *args, **kw_args)

bind_point(Channel)

bind_routine(threaded_object, lifecycle=False, message_trail=False, execution_trace=False, user_logs=USER_LOG.NONE)
bind_routine(object_dispatch, lifecycle=False, message_trail=False, execution_trace=False, user_logs=USER_LOG.NONE)



# test_publisher.py
import kipjak as kj

from test_api import *
from test_function import *

# A bare-bones implementation of a published network service.
DEFAULT_NAME = 'acme'

def publisher(self, name: str=None, scope: kj.ScopeOfDirectory=None):
	'''Establish a named service, wait for clients and their enquiries. Return nothing.'''
	name = name or DEFAULT_NAME
	scope = scope or kj.ScopeOfDirectory.GROUP

	kj.publish(self, name, scope=scope)		# Register this object under the given name.

	m = self.input()
	if isinstance(m, kj.Published):		# Name registered with directory.
		pass

	elif isinstance(m, kj.Faulted):		# Any fault, e.g. NotPublished.
		self.complete(m)

	# Run a live directory service. Framework notifications and
	# client requests.
	while True:
		m = self.input()
		if isinstance(m, (kj.Delivered, kj.Dropped)):	# Subscribers coming and going.
			continue

		elif isinstance(m, Xy):
			t = texture(self, x=m.x, y=m.y)
			m = kj.cast_to(t, table_type)
			self.reply(m)					# Respond to client.

		elif isinstance(m, kj.Stop):
			self.complete(kj.Aborted())

kj.bind(publisher)				# Register with framework.

if __name__ == '__main__':		# Process entry-point.
	kj.create(publisher)



# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Logging methods available within the async runtime.

"""
__docformat__ = 'restructuredtext'

from .virtual_point import *
from .virtual_runtime import *
from .point_runtime import *
from .point_machine import *
from .bind_type import *

__all__ = [
	'PEAK_BEFORE_BLOCKING',
	'LogAgent',
]

#
#
PEAK_BEFORE_BLOCKING = 1024 * 64

class LogAgent(Threaded, Stateless):
	"""
	A stateless, async object that accepts PointLog messages originating
	from calls to ``Point.`` ``debug``, ``trace``, ``console``,
	``warning`` and ``fault``, and presents them to a saved method that
	might be the default ``log_to_stderr`` or an application-defined
	method.

	The ``LogAgent`` derives from ``Queue``, i.e. it will enjoy its
	own dedicated thread. The result is that logs can originate from
	any runtime **kipjak** thread and will be collected/serialized at the
	sole LogAgent instance. The saved log method can enjoy the
	knowledge that all multi-threading issues are resolved and that
	logs are buffered in a reliable and robust fashion.

	A part of delivering on those promises is the custom initialization
	of the underlying Queue object. It is set to blocking of upstream
	sources and the size of the Queue is set to a generous, custom
	value.

	Note the complete disabling of all logging for ``LogAgent``. It
	doesnt make much sense for the logger to log to itself and
	certainly not during creation where at some point it exists
	but is not yet entered into **kipjak** tables. At such a moment
	send would fail - at best.
	"""

	def __init__(self, method):
		Threaded.__init__(self, blocking=True, maximum_size=PEAK_BEFORE_BLOCKING)
		Stateless.__init__(self)
		self.method = method
		self.tap = []

def LogAgent_Start(self, message):
	pass

def LogAgent_PointLog(self, message):
	try:
		line = self.method(message)
	except Exception as e:
		s = str(e)
		return

def LogAgent_RedirectLog(self, message):
	redirect = message.redirect
	try:
		redirect.from_previous(self.method)
	except AttributeError:
		pass
	self.method = redirect

def LogAgent_OpenTap(self, message):
	self.tap.append(self.return_address)

def LogAgent_CloseTap(self, message):
	try:
		self.tap.remove(self.return_address)
	except ValueError:
		pass

def LogAgent_Enquiry(self, message):
	"""
	A query from the ``tear_down`` function to ensure that all previous
	messages have been processed.
	"""
	self.reply(Ack())

def LogAgent_Stop(self, message):
	self.complete()

def LogAgent_int(self, message):
	self.complete()

def LogAgent_float(self, message):
	self.complete()

def LogAgent_dict_str_int(self, message):
	self.complete()

def LogAgent_Stop(self, message):
	self.complete()

LOG_AGENT_DISPATCH = (Start,
	PointLog,
	Enquiry, Stop,
	int,
	float,
	dict[str,int])

bind_stateless(LogAgent, dispatch=LOG_AGENT_DISPATCH,
	lifecycle=False, message_trail=False,
	execution_trace=False, user_logs=USER_LOG.NONE)
